import * as crypto from 'crypto';
import { WsCloseFrame, WsFrameDecoder, WsFrameHeader, WsPingFrame, WsPongFrame } from '../codec';
import { utf8Size } from '../../../../util/strings/utf8';
export class WsServerConnection {
    encoder;
    socket;
    closed = false;
    maxIncomingMessage = 2 * 1024 * 1024;
    maxBackpressure = 2 * 1024 * 1024;
    stream = null;
    defaultOnPing = (data) => {
        this.sendPong(data);
    };
    onmessage = () => { };
    onping = this.defaultOnPing;
    onpong = () => { };
    onclose = () => { };
    constructor(encoder, socket) {
        this.encoder = encoder;
        this.socket = socket;
        const decoder = new WsFrameDecoder();
        let currentFrame = null;
        const handleData = (data) => {
            try {
                decoder.push(data);
                if (currentFrame) {
                    const length = currentFrame.length;
                    if (length <= decoder.reader.size()) {
                        const buf = new Uint8Array(length);
                        decoder.copyFrameData(currentFrame, buf, 0);
                        const isText = currentFrame.opcode === 1;
                        currentFrame = null;
                        this.onmessage(buf, isText);
                    }
                }
                while (true) {
                    const frame = decoder.readFrameHeader();
                    if (!frame)
                        break;
                    else if (frame instanceof WsPingFrame)
                        this.onping(frame.data);
                    else if (frame instanceof WsPongFrame)
                        this.onpong(frame.data);
                    else if (frame instanceof WsCloseFrame)
                        this.onClose(frame.code, frame.reason);
                    else if (frame instanceof WsFrameHeader) {
                        if (this.stream) {
                            if (frame.opcode !== 0) {
                                this.onClose(1002, 'DATA');
                                return;
                            }
                            throw new Error('streaming not implemented');
                        }
                        const length = frame.length;
                        if (length > this.maxIncomingMessage) {
                            this.onClose(1009, 'TOO_LARGE');
                            return;
                        }
                        if (length <= decoder.reader.size()) {
                            const buf = new Uint8Array(length);
                            decoder.copyFrameData(frame, buf, 0);
                            const isText = frame.opcode === 1;
                            this.onmessage(buf, isText);
                        }
                        else {
                            currentFrame = frame;
                        }
                    }
                }
            }
            catch (error) {
                this.onClose(1002, 'DATA');
            }
        };
        const handleClose = (hadError) => {
            if (this.closed)
                return;
            this.onClose(hadError ? 1001 : 1002, 'END');
        };
        socket.on('data', handleData);
        socket.on('close', handleClose);
    }
    close() {
        const code = 1000;
        const reason = 'CLOSE';
        const frame = this.encoder.encodeClose(reason, code);
        this.socket.write(frame);
        this.onClose(code, reason);
    }
    onClose(code, reason) {
        this.closed = true;
        if (this.__writeTimer) {
            clearImmediate(this.__writeTimer);
            this.__writeTimer = null;
        }
        const socket = this.socket;
        socket.removeAllListeners();
        if (!socket.destroyed)
            socket.destroy();
        this.onclose(code, reason);
    }
    upgrade(secWebSocketKey, secWebSocketProtocol, secWebSocketExtensions) {
        const accept = secWebSocketKey + '258EAFA5-E914-47DA-95CA-C5AB0DC85B11';
        const acceptSha1 = crypto.createHash('sha1').update(accept).digest('base64');
        this.socket.write('HTTP/1.1 101 Switching Protocols\r\n' +
            'Upgrade: websocket\r\n' +
            'Connection: Upgrade\r\n' +
            'Sec-WebSocket-Accept: ' + acceptSha1 + '\r\n' +
            (secWebSocketProtocol ? 'Sec-WebSocket-Protocol: ' + secWebSocketProtocol + '\r\n' : '') +
            '\r\n');
    }
    __buffer = [];
    __writeTimer = null;
    write(buf) {
        if (this.closed)
            return;
        this.__buffer.push(buf);
        if (this.__writeTimer)
            return;
        this.__writeTimer = setImmediate(() => {
            this.__writeTimer = null;
            const buffer = this.__buffer;
            this.__buffer = [];
            if (!buffer.length)
                return;
            const socket = this.socket;
            if (socket.writableLength > this.maxBackpressure)
                this.onClose(1009, 'BACKPRESSURE');
            socket.cork();
            for (let i = 0, len = buffer.length; i < len; i++)
                socket.write(buffer[i]);
            socket.uncork();
        });
    }
    sendPing(data) {
        const frame = this.encoder.encodePing(data);
        this.write(frame);
    }
    sendPong(data) {
        const frame = this.encoder.encodePong(data);
        this.write(frame);
    }
    sendBinMsg(data) {
        const encoder = this.encoder;
        const header = encoder.encodeDataMsgHdrFast(data.length);
        this.write(header);
        this.write(data);
    }
    sendTxtMsg(txt) {
        const encoder = this.encoder;
        const writer = encoder.writer;
        const size = utf8Size(txt);
        encoder.writeHdr(1, 1, size, 0);
        writer.ensureCapacity(size);
        writer.utf8(txt);
        const buf = writer.flush();
        this.write(buf);
    }
}
